from scrapy.contrib.spiders import CrawlSpider, Rule
from scrapy.contrib.linkextractors.sgml import SgmlLinkExtractor
from scrapy.selector import Selector
from scrapy.item import Item

import bc_validator as bcv
from bitcoin.items import BitcoinItem

class BitcointalkSpider(CrawlSpider):
    name = "bitcointalk"
    allowed_domains = ["bitcointalk.org"]
    
    start_urls = ["https://bitcointalk.org/index.php"]

    rules = (
        Rule(SgmlLinkExtractor(deny=[
            'https://bitcointalk\.org/index\.php\?action=ignore',
            'https://bitcointalk\.org/index\.php\?action=profile',
            ], 
            allow_domains='bitcointalk.org'), callback='parse_item', follow=True),
    )

    def parse_item(self, response):
        sel = Selector(response)
        sites = sel.xpath('//tr[contains(@class, "td_headerandpost")]')
        items = []
        for site in sites:
            item = BitcoinItem()
            item["membername"] = site.xpath('.//td[@class="poster_info"]/b/a/text()').extract()
            addresses = site.xpath('.//div[contains(@class, "signature")]/text()').re(r'(1[1-9A-HJ-NP-Za-km-z]{26,33})')
            if item["membername"] and addresses:
                addr_list = set()
                for addr in addresses:
                    if (bcv.check_bc(addr)):
                        addr_list.add(addr)
                item["address"] = addr_list
                if len(addr_list) > 0:
                    items.append(item)
        return items
