# -*- coding: utf-8 -*-
# <nbformat>2</nbformat>

# <codecell>

import time, timeit, json, http_server

import numpy as np
import networkx as nx
import pandas as pd
from collections import defaultdict

import pygephi
from networkx.readwrite import json_graph
from networkx.readwrite import write_gexf

from pandasql import sqldf
from pandasql import load_meat, load_births
pysqldf = lambda q: sqldf(q, globals())

from bitcoin_transaction_network_parser import read_uic_data

# <codecell>

# Read the data in uic format
uic_data = read_uic_data('/home/<username>/bitcoin/data/processed-silkroad-arrest-oct-2013',None)
pub_keys, xn_keys, user_edges, user_edge_inputs, \
	  user_edge_inputs_public_keys, userkey_list = uic_data
print user_edges[:10]

# <codecell>

distinct_user_edges = user_edges #[user_edges['user_from_key'] != 1]
print len(distinct_user_edges), 'out of', len(user_edges)

# <codecell>

# distinct_user_edges = user_edges[user_edges['user_from_key'] != 1]
pysqldf("""SELECT * FROM user_edges e """
	"""WHERE e.user_to_key == '18145' """)

# <codecell>

# Cross-referencing bitcointalk.org
import json
crawled = json.loads(open('/home/<username>/bitcoin/analysis/bitcoin/bitcoin/crawled-output-2013-12-09.json').read())
# print crawled.values()[:2]

bitcointalk_users = {}
for user,pks in crawled.iteritems(): 
    for pk in pks: 
	try : 
	    uid = pk_search.query(str(pk))
	    if uid is not None: 
		print pk, 'Found user', user, uid, str(uid) in top_nodes, str(uid) in G
		node_attrs = { 'size':50, 'r':1.0, 'g':0.0, 'b':0.0, \
			      'label':str(user) }
		bitcointalk_users[str(uid)] = node_attrs
		print 'Found attrs: ', node_attrs
	except UnicodeEncodeError: 
	    print pk

print bitcointalk_users

# <codecell>

# Start up gephi client
g = pygephi.GephiClient('http://localhost:8080/workspace0', autoflush=True)

# <codecell>

# k-th order neighbors of a starting node
def knbrs(G, start, k): 
    nbrs = set([start])
    for l in range(k):
        nbrs = set((nbr for n in nbrs for nbr in G[n]))
    return nbrs

# Construct directed graph of all edges 
t = time.time()
G = nx.DiGraph()
G.add_edges_from(zip(distinct_user_edges['user_from_key'].astype(str), \
		     distinct_user_edges['user_key_to'].astype(str)))
print 'Construct Graph: %i nodes' % len(G.nodes())
print 'Construct Graph: %i edges' % len(G.edges())    
print 'Constructing di-graph took ', time.time() - t, 's\n'

# Compute page rank on the full graph
t = time.time()
pr_nodes = nx.pagerank_scipy(G, alpha=0.85, weight=None)
print 'Page rank took ', time.time() - t, 's\n'

# Get the top PageRank nodes
t = time.time()
top_pr_nodes_ = sorted(zip(pr_nodes.keys(), pr_nodes.values()), key=lambda x: x[1], reverse=True)[:10];
print 'TOP Page Rank Nodes: ', top_pr_nodes_[:10]
print 'Sorting pageranked nodes took ', time.time() - t, 's\n'

 
# Find the 1st order nodes of the top PageRank nodes
t = time.time()
top_pr_nodes = []
for node in top_pr_nodes_: 
    # order = 1
    nbrs = knbrs(G, node[0], 1)
    top_pr_nodes.extend([ (nbr, pr_nodes[nbr]) for nbr in nbrs ])
print 'Finding 1st order nodes took ', time.time() - t, 's\n'

# <codecell>

# Visualize nodes
# Clean up gephi
g.clean()

# Add nodes to gephi 
ids_with_attrs = {}
pos = np.random.random((len(top_pr_nodes),2)) * 5000
for idx,node_pr in enumerate(top_pr_nodes): 
    def_attr = { 'size':10, 'r':0.0, 'g':0.0, 'b':1.0, \
		 'x':pos[idx][0], 'y':pos[idx][1], \
		 'label':node_pr[0], 'weight':node_pr[1] }

    if node_pr[0] in bitcointalk_users.keys(): 
	node_attrs = bitcointalk_users[node_pr[0]]
    else: 
	node_attrs = def_attr
    ids_with_attrs[node_pr[0]] = node_attrs

    # store for future updates, cross-referencing
    G.node[node_pr[0]] = node_attrs

g.add_nodes(ids_with_attrs)

top_nodes = [ node for node,pr in top_pr_nodes ]
all_edges = zip(distinct_user_edges['user_from_key'].astype(str), \
		distinct_user_edges['user_key_to'].astype(str), \
		str(distinct_user_edges.index.tolist()), \
		# distinct_user_edges['date'].to_string(), \
		distinct_user_edges['value'].astype(str))

# Only add edges between top nodes
top_pr_edges = []
for edge in all_edges: 
    # If the node is involved in the edge (either from or to or both)
    if edge[0] in top_nodes or edge[1] in top_nodes: 
	top_pr_edges.append(edge)

# Add edges to gephi
ids_with_attrs = {}
for idx,edge in enumerate(top_pr_edges): 
    ids_with_attrs[idx] = {'directed':True, \
			   'source':edge[0], 'target':edge[1], \
			   'date':edge[2], 'weight':edge[3]}
    # g.add_edge(idx,edge[0],edge[1],edge_attrs)
g.add_edges(ids_with_attrs)

# <markdowncell>

# node_ids_with_attrs = dict([(node_ids_with_attrs[node],G.node[node]) 
# 			    for node in G.nodes()])
# edge_ids_with_attrs = dict([(node_ids_with_attrs[node],G.node[node]) 
# 			    for edges in G.edges()])

# <codecell>

# Public Key Index Search (pubkey->userid)
class PubKeySearch: 
    # Build inverted pubkey index (pubkey->userid) 
    # WARNING: 1-based indexing by the
    # uic code makes it complicated to index to the user correctly
    def __init__(self, pub_keys): 
	t = time.time()

	all_pks, all_vals = [], []
	for idx,userkey in enumerate(userkey_list): 
	    pks = pub_keys.ix[userkey]['public_key_string'].values
	    all_pks.extend(pks)
	    
	    # +1 for (None), and +1 for 1-based indexing
	    all_vals.extend([idx + 2] * len(pks)) 

	self.pubkey_to_user = dict(zip(all_pks,all_vals))
	print 'Constructing inverted pubkey index took ', time.time() - t, 's\n'

    def query(self, pk): 
	if pk in self.pubkey_to_user: 
	    return self.pubkey_to_user[pk]
	return None;


# Search by Public Key
FBI_pk = '1FfmbHfnpaZjKFvyi1okTjJJusN455paPH'
pk_search = PubKeySearch(pub_keys);
print 'FBI user id: ', pk_search.query(FBI_pk)

# Search by node id
print 'Search by node id: '
print pub_keys.ix[userkey_list[9103-2]]['public_key_string']

# <markdowncell>

# # Clean up gephi
# g.clean()
# 
# # Add nodes to gephi 
# ids_with_attrs = {}
# pos = np.random.random((len(G.nodes()),2)) * 5000
# for idx,node in enumerate(G.nodes()): 
#     ids_with_attrs[node] = {'size':10, 'r':0.0, 'g':0.0, 'b':1.0, \
# 			   'x':pos[idx][0], 'y':pos[idx][1], \
# 			   'label':node, 'weight':1 #len(userkey_list[int(node)-1])
# 		       }
# g.add_nodes(ids_with_attrs)
# 
# pruned_edges = zip(distinct_user_edges['user_from_key'][:NROWS].astype(str), \
# 		   distinct_user_edges['user_key_to'][:NROWS].astype(str), \
# 		   distinct_user_edges['date'][:NROWS].to_string(), \
# 		   distinct_user_edges['value'][:NROWS].astype(str))
# 
# # Add edges to gephi
# ids_with_attrs = {}
# for idx,edge in enumerate(pruned_edges): 
#     ids_with_attrs[idx] = {'directed':True, \
# 			   'source':edge[0], 'target':edge[1], \
# 			   'date':edge[2], 'weight':edge[3]}
#     # g.add_edge(idx,edge[0],edge[1],edge_attrs)
# g.add_edges(ids_with_attrs)
# 
# # # Adding other sources 
# # node_attrs = {"size":20, 'r':1.0, 'g':0.0, 'b':0.0, 'x':20, 'y':20}
# # pos = np.random.random((10,2)) * 1000
# # for j in range(0,10): 
# #     node_attrs['wikileaks'] = node
# #     node_attrs['x'] = pos[j][0]
# #     node_attrs['y'] = pos[j][1]
# #     g.add_node('wikileaks-%i' % j, **node_attrs)

# <markdowncell>

# # pr_list = zip(pr.keys(), pr.values()); 
# # pr_list_sorted = sorted(pr_list, key=lambda x: x[1], reverse=True)[:20]
# # keys,score = zip(*pr_list_sorted)
# 
# for n in G: 
#     G.node[n]['name'] = n
# # d = json_graph.node_link_data(G)
# # f = open('viz/graph.json','w')
# # json.dump(d, f)
# # f.close()
# write_gexf(G, 'viz/graph.gexf')
# print('Wrote node-link JSON data to viz/graph.gexf')
# # open URL in running web browser
# # http_server.load_url('force/force.html')
# # print('Or copy all files in force/ to webserver and load force/force.html')

# <markdowncell>

# distinct = pysqldf("""SELECT * FROM user_edges e WHERE e.user_from_key == user_key_to""")
# print distinct[:10]

# <markdowncell>

# distinct_user_edges = user_edges[user_edges['user_from_key'] == user_edges['user_key_to']]
# 
# print len(user_edges), len(distinct_user_edges)
# print distinct_user_edges[:10]

# <markdowncell>

# meat = load_meat()
# births = load_births()
# 
# q = """SELECT m.date, m.beef, b.births FROM meat m LEFT JOIN births b ON m.date = b.date WHERE m.date > '1974-12-31';"""
# df = pysqldf(q)

