from optparse import OptionParser
from datetime import timedelta
from datetime import date
import bitcoin_transaction_network_parser as btnp
import pandas as pd
import datetime
import numpy as np
import time
import os
import csv
import pickle
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import matplotlib

#====globals===
data_dir = 'F:/bitcoinData/processed-2012-2013'

class BtToUSD:
    """
        Class for converting from Bt to USD on a given day
    """
    def __init__(self):
        self.prices = pd.read_csv('data/market-price_days.txt',
                                  names=['day', 'usd'],
                                  index_col=[0],
                                  dtype={'day': datetime.date, 'usd': np.float32},
                                  dayfirst=True,
                                  parse_dates=[0])

    def getUSD(self, btValue, day):
        """
        Parameters
        ---------
        btValue : bitcoin value to convert to USD
        day : day for which we want to perform the conversion

        Returns:
        ------
        btValue * conversion-rate from the given day
        we use prices taken from : https://blockchain.info/charts/market-price
        """
        if type(day) != pd.Timestamp:
            raise Exception("expected pandas timestamp in getUSD")
        delta = datetime.timedelta(hours=day.hour, minutes=day.minute,
                                   seconds=day.second, microseconds=day.microsecond)
        return self.prices.ix[day - delta]['usd'] * btValue

    def getBt(self, usd, day):
        """
        Parameters
        ---------
        usd : USD value to convert to bt
        day : day for which we want to perform the conversion

        Returns:
        ------
        usd/(usd/bt conversion-rate from the given day)
        we use prices taken from : https://blockchain.info/charts/market-price
        """
        return usd/self.getUSD(1, day)


def subset(df, columnName, start, end, startIndex=None):
    """
    Returns df subset where the column values are between start and end

    Parameters
    ----------
    df : pandas.DataFrame with the given column name
    start : of same type as the column
    end - of same type as the column
    """
    if columnName != df.index.name:
        if startIndex is not None:
            raise Exception("startIndex should be none for non-index col")
        result = df.ix[df[columnName] >= start]
        return result.ix[result[columnName] <= end]
    else:
        result = df[df.index >= start]
        return result[result.index <= end]
    """
    if startIndex is None:
        raise Exception("should provide a start index")
    sInd = startIndex
    for i in xrange(startIndex, -1, -1):
        if df.index[i] >= start:
            sInd = i
        else:
            break
    eInd = startIndex
    for i in xrange(startIndex, len(df), 1):
        if df.index[i] <= end:
            eInd = i
        else:
            break
    return df.iloc[range(sInd, eInd+1)]
    """


def parse(just_edges=True):
    """
    Parameters
    ----------
    data_dir : directory w/ data files
    just_edges : bool , true if only want user_edges.txt
    """
    parses = OptionParser()
    parses.add_option("-n", "--nrows", dest="nrows",
                      help="Input Number of rows to process",
                      default=1e4,
                      metavar="NROWS")
    parses.add_option("-d", "--data-dir", dest="data_dir",
                      help="Data directory where all the files (pubkey_list.txt etc. ) are located",
                      metavar="NROWS")
    (options, args) = parses.parse_args()

    # Required options =================================================
    if not options.data_dir:
        #parses.error('Data Directory not given')
        #options.data_dir = 'F:/bitcoinData/bitcoin_uic_data'
        options.data_dir = data_dir

    pub_keys, xn_keys, user_edges, user_edge_inputs, user_edge_inputs_public_keys, userkey_list = \
        btnp.read_uic_data(options.data_dir, options.nrows, edges_only=just_edges)

    '''
    print '\n\n  Public key list ============================================'
    print pub_keys[:10]
    print '\n\n  Transaction key list ======================================='
    print xn_keys[:10]
    print '\n\n  User Edges ================================================='
    print user_edges[:10]
    print '\n\n User edge inputs ==========================================='
    print user_edge_inputs
    print '\n\n Input public keys =========================================='
    print user_edge_inputs_public_keys
    print '\n\n user key components ========================================'
    print userkey_list
    '''
    return user_edges


def avgCollisions(edges, timediff=timedelta(minutes=10), usdDiff=5):
    """
    Paramaters
    -----------
    usdDiff : float, difference in usd
    timediff : timedelta, time difference window
    """
    method_start_time = time.time()

    numOverlapping = 0.0
    btUsdConverter = BtToUSD()
    for i in xrange(len(edges)):
        #compute time neighborhood
        t = edges.index[i]
        twindow = subset(edges, 'date', t - timediff, t + timediff, i)

        #compute intersection with value neighborhood
        v = edges.ix[i]['value']
        coinDiff = btUsdConverter.getBt(usdDiff, t)
        tAndValDiff = subset(twindow, 'value', v - coinDiff, v + coinDiff)
        neighbors = len(tAndValDiff) - 1
        numOverlapping += neighbors
        if i % 10000 == 0:
            print 'getSimilar processed ', '{:,}'.format(i), '/', '{:,}'.format(len(edges)), ' rows'

    print 'numOverlapping = ', numOverlapping
    print 'getSimilar took %3.2f s' % (time.time() - method_start_time)
    return numOverlapping/float(len(edges))


def splitEdgesByMonth(input_directory=data_dir, output_directory='C:/bitcoinData/splits'):
    """
    Splits the user_edges.txt file into separate files for each month
    """
    #transaction_key,
    #user_from_key,
    #user_key_to,
    #date,
    # value
    file = ''.join([input_directory, '/user_edges.txt'])
    #output_directory = input_directory + '/splits'
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    files_writing = {}
    num_lines_processed = 0
    with open(file, 'r') as f:
        reader = csv.reader(f, delimiter=',')
        for row in reader:
            #compute first of month
            t = pd.Timestamp(row[3])
            first_of_month = datetime.date(year=t.year, month=t.month, day=1)
            out_file = output_directory + '/' + first_of_month.strftime('%Y_%m') + '.txt'

            #see if we have an open writer for this
            if not out_file in files_writing:
                files_writing[out_file] = open(out_file, 'a')

            #append
            files_writing[out_file].write(','.join(row) + '\n')

            if num_lines_processed % 1e5 == 0:
                print 'processed ', '{:,}'.format(num_lines_processed), ' lines'
            num_lines_processed += 1

    for f in files_writing.values():
        f.close()


def iterateEdgesByMonth(splits_dir=data_dir+'/splits'):
    """
    -----
    Returns : dataframe, file_name for each month
    """
    dtype_dict = {'transaction_key': np.int32,
                  'user_from_key': np.int32,
                  'user_key_to': np.int32,
                  'value': np.float32}

    for f in os.listdir(splits_dir):
        if not os.path.isfile(os.path.join(splits_dir, f)):
            continue
        print 'reading ', f
        read_start_time = time.time()
        user_edges = pd.io.parsers.read_csv(os.path.join(splits_dir, f),  # filename
                                            names=["transaction_key", "user_from_key",
                                                   "user_key_to", "date", "value"],
                                            dtype=dtype_dict,
                                            index_col=3, parse_dates=[3])
        print 'reading took %3.2f s' % (time.time() - read_start_time)
        yield user_edges, f


def computeUniquenessByMonth(splits_dir=data_dir+'/splits'):
    minutes_deltas = [5, 15]  # [5, 15, 25]
    usd_deltas = [1, 10]  # [1, 10, 100]

    out_dir = splits_dir + '/stats'
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    for user_edges, f in iterateEdgesByMonth():
        print 'num_lines = ', '{:,}'.format(len(user_edges)), '\n'
        if not '.txt' in f:
            raise Exception("should have been txt file")

        #compute
        stats = {}
        for minDiff in minutes_deltas:
            for usdDiff in usd_deltas:
                print '[computing] %s :  minutes_diff = %d : usd_diff = %d' % (f, minDiff, usdDiff)
                stats[(minDiff, usdDiff)] = avgCollisions(user_edges,
                                                          timediff=timedelta(minutes=minDiff),
                                                          usdDiff=usdDiff)
                print '\nstats so far:\n', stats

        #write to disk
        out_file = os.path.join(out_dir, f.replace('.txt', '_stats.txt'))
        if os.path.exists(out_file):
            raise Exception('why does this file already exist?: ' + out_file)
        print 'writing pickle'
        pickle.dump(stats, open(out_file, 'wb'))
        print 'wrote pickle\n\n\n==================\n'


def plot(stats_dir=data_dir+'/splits/stats'):
    #===read in data
    lines = {}
    x_axis = []
    for f in os.listdir(stats_dir):
        year_month = f.replace("_stats.bin", "").split('_')
        next_date = date(year=int(year_month[0]), month=int(year_month[1]), day=1)
        x_axis.append(next_date)
        fname = os.path.join(stats_dir, f)
        if not os.path.isfile(fname):
            continue
        print 'reading ', fname
        d = pickle.load(open(fname, 'rb'))
        for key in d.keys():
            if not key in lines:
                lines[key] = []
            lines[key].append(d[key])
    print '\n\n\n=============\n'

    #===plotting
    fig, ax = plt.subplots(1)
    fig.autofmt_xdate()
    ax.xaxis.fmt_xdata = mdates.DateFormatter('%Y-%m')
    plt.xlabel('date', fontsize=25)
    plt.ylabel('avg # matching transactions', fontsize=25)
    for key, value in lines.items():
        ax.plot(x_axis, value, '-', linewidth=2, label='%d minutes, $%d' % key)

    leg = ax.legend(ncol=2, shadow=True, title="(minute, time) windows", prop={'size': 25})
    leg.get_title().set_fontsize('25')
    plt.title('Transaction matches within ($,Time) Windows', fontsize=30)
    plt.show()


if __name__ == "__main__":
    #print getSimilar()
    #computeUniquenessByMonth()
    plot()

    """
    y = [[7, 10, 9, 1],
         [1, 2, 3, 4]]
    dates = [datetime.date(year=2012, month=i, day=1) for i in range(1, 1 + len(y[0]))]
    fig, ax = plt.subplots(1)
    fig.autofmt_xdate()
    ax.fmt_xdata = mdates.DateFormatter('%Y-%m')
    ax.AutoMinorLocator = plt.NullLocator()
    line1 = ax.plot(dates, y[0], '-', linewidth=2, label='first thing')
    line2 = ax.plot(dates, y[1], '-', linewidth=2, label='2nd thing')
    ax.legend(ncol=2, shadow=True, title="Legend")
    plt.xlabel('date')
    plt.ylabel('avg_collisions / transaction')
    plt.show()
    """