import networkx as nx
import time
import json
import bitcoin_transaction_network_parser as btnp
from bitcoin_user_search_utils import PubKeySearch
import os


def shortest_path(edges, start_set, end_set):
    """
    Finds the shortest path between 2 sets of nodes

    Parameters
    ----------
    edges : dataframe, the graph
    start_set : collection<int> - the ints are user id's
    end_set : collection<int> the ints are user id's
    """
    start_time = time.time()
    G = nx.DiGraph()

    edges_list = zip(edges['user_from_key'].astype(str),
                     edges['user_key_to'].astype(str))

    pruned_edges = []
    for edge in edges_list: 
        if edge[0] == '0' or edge[1] == '0': continue
        pruned_edges.append(edge)
    G.add_edges_from(pruned_edges)

    print 'Constructed Graph with %i nodes, %i edges' % (len(G.nodes()), len(G.edges()))
    print 'took time ', time.time() - start_time, ' secs'
    best_path = None

    for next_start in start_set:
        for next_end in end_set:
            if next_start not in G.nodes() or next_end not in G.nodes(): 
                continue
            try:
                p = nx.shortest_path(G,
                                     source=str(next_start),
                                     target=str(next_end))
            except nx.exception.NetworkXNoPath:
                continue
            if best_path is None or len(p) < len(best_path):
                best_path = p
    return best_path


def read_scraped_data(known_entities_file=None):
    """
    Read user-id's from json file of scraped data

    Parameters
    --------
        known_entities_file : string, json absolute filename
    """
    if known_entities_file is None:
        known_entities_file = os.path.join(os.getcwd(), os.pardir, 'bitcoin/bitcoin/crawled-output-2013-12-10-pretty.json')
        print 'using known_entities_file = ', known_entities_file

    stuff = json.load(open(known_entities_file, 'r'))
    print 'Crawled Users: ', len(stuff)

    # skipped = 0
    # for key in stuff.keys():
    #     cleaned = []
    #     try:
    #         for i in stuff[key]:
    #             cleaned.append(str(i).strip())
    #         print key, ' Cleaned: ', cleaned
    #         stuff[key] = cleaned
    #     except:
    #         skipped += 1
    #         pass

    # print 'Skipped: ', skipped

    #==========now read crypto_lock
    '''
    un = 'cryptolocker'
    stuff[un] = []
    for line in open('data/cryptoLockerAddys.txt'):
        split = line.split()
        if len(split) == 2:
            stuff[un].append(str(split[0]).strip())
        if len(split) == 5:
            stuff[un].append(str(split[2]).strip())

    print stuff
    '''
    return stuff


if __name__ == "__main__":

    known_entities = read_scraped_data()

    #======read edges, keys
    pub_keys, xn_keys, user_edges, user_edge_inputs, user_edge_inputs_public_keys, userkey_list = \
        btnp.read_uic_data('data/processed-silkroad-arrest-oct-25-2013')

    #determine silk road indices
    inds = user_edges['value'] == 324
    silk_road = set(user_edges[inds]['user_from_key'].values)
    silk_road.remove(0)
    print 'Number of silk road entities: %i ' % len(silk_road)
    
    #===indexing
    pkInd = PubKeySearch(pub_keys=pub_keys,
                         userkey_list=userkey_list)

    # Determine user id for wikileaks
    wikileaks = '1HB5XMLmzFVj8ALj6mfBsbifRoD4miY36v'
    wikileaks_ind = pkInd.query(wikileaks)
    if wikileaks_ind is not None: silk_road.add(wikileaks_ind)
    print 'Found wikileaks: ', wikileaks_ind is not None

    #======setup search
    from_ids = set([])
    for user_name, keys in known_entities.items():
        for pk in keys:
            user_id = pkInd.query(pk)
            if user_id is None:
                continue  # pk not in our dataset
            print 'user_name=', user_name, '\tpk=', pk, '\tuser_id=', user_id
            from_ids.add(user_id)

    # Remove clumped group
    from_ids.remove(0)

    print 'from ids: ', from_ids
    fbi_id = 16423
    print shortest_path(user_edges, from_ids, silk_road)

